;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "consult-project-extra" "20260428.2007"
  "Consult integration for project.el."
  '((emacs   "28.1")
    (consult "0.17")
    (project "0.8.1"))
  :url "https://github.com/Qkessler/consult-project-extra"
  :commit "52c453b7f85dea90cc53ff8fba750795606415df"
  :revdesc "52c453b7f85d"
  :keywords '("convenience" "project" "management"))
